
typedef enum TRBType {
  TRB_RESERVED = 0,
  TR_NORMAL,
  TR_SETUP,
  TR_DATA,
  TR_STATUS,
  TR_ISOCH,
  TR_LINK,  // 6
  TR_EVDATA,
  TR_NOOP,
  CR_ENABLE_SLOT,         // 9
  CR_DISABLE_SLOT,        // 10
  CR_ADDRESS_DEVICE,      // 11
  CR_CONFIGURE_ENDPOINT,  // 12
  CR_EVALUATE_CONTEXT,    // 13
  CR_RESET_ENDPOINT,      // 14
  CR_STOP_ENDPOINT,       // 15
  CR_SET_TR_DEQUEUE,      // 16
  CR_RESET_DEVICE,        // 17
  CR_FORCE_EVENT,
  CR_NEGOTIATE_BW,
  CR_SET_LATENCY_TOLERANCE,
  CR_GET_PORT_BANDWIDTH,  // 21
  CR_FORCE_HEADER,
  CR_NOOP,
  ER_TRANSFER = 32,
  ER_COMMAND_COMPLETE,
  ER_PORT_STATUS_CHANGE,
  ER_BANDWIDTH_REQUEST,
  ER_DOORBELL,
  ER_HOST_CONTROLLER,
  ER_DEVICE_NOTIFICATION,
  ER_MFINDEX_WRAP,
  /* vendor specific bits */
  CR_VENDOR_NEC_FIRMWARE_REVISION = 49,   // 49
  CR_VENDOR_NEC_CHALLENGE_RESPONSE = 50,  // 50

  DOOR_BELL_WRITE,
} TRBType;

typedef struct XHCITRB {
  uint64_t parameter;
  uint32_t status;
  uint32_t control;
} XHCITRB;

typedef struct Parameter {
    uint32_t ictl_ctx[2];
    uint8_t padding[24];
    uint32_t slot_ctx[4];
} Parameter;

typedef struct XHCIEvRingSeg {
  uint32_t addr_low;
  uint32_t addr_high;
  uint32_t size;
  uint32_t rsvd;
} XHCIEvRingSeg;


#define OFF_OPER        0x40
#define OFF_RUNTIME     0x1000
#define OFF_DOORBELL    0x2000
#define OFF_MSIX_TABLE  0x3000
#define OFF_MSIX_PBA    0x3800

/* 5.3 Host Controller Capability Registers
 * Table 19
 */
struct xhci_cap_regs {
  uint8_t caplength;
  uint8_t reserved;
  uint16_t hciversion;
  uint32_t hcsparams1;
  uint32_t hcsparams2;
  uint32_t hcsparams3;
  uint32_t hccparams;
#define XHCI_HCCPARAMS_CSZ BIT(2)
#define XHCI_HCCPARAMS_XECP(x) ((x & 0xFFFF0000) >> 16)
  uint32_t dboff;
  uint32_t rtsoff;
} __attribute__((packed));

/* Table 27: Host Controller USB Port Register Set */
struct xhci_port_regs {
  uint32_t portsc;
#define PORTSC_CCS BIT(0)
#define PORTSC_PED BIT(1)
#define PORTSC_OCA BIT(3)
#define PORTSC_PR BIT(4)
#define PORTSC_PLS_MASK (0xF << 5)
#define PORTSC_PLS_U0 0
#define PORTSC_PLS_U1 1
#define PORTSC_PLS_U2 2
#define PORTSC_PLS_U3 3
#define PORTSC_PLS_DISABLED 4
#define PORTSC_PLS_RXDETECT 5
#define PORTSC_PLS_INACTIVE 6
#define PORTSC_PLS_POLLING 7
#define PORTSC_PLS_RECOVERY 8
#define PORTSC_PLS_HOTRESET 9
#define PORTSC_PLS_COMP_MODE 10
#define PORTSC_PLS_TEST_MODE 11
#define PORTSC_PLS_RESUME 15
#define PORTSC_PP BIT(9)
#define PORTSC_PS_MASK (0xF << 10)
#define PORTSC_PIC_MASK (0x3 << 14)
#define PORTSC_LWS BIT(16)
#define PORTSC_CSC BIT(17)
#define PORTSC_PEC BIT(18)
#define PORTSC_WRC BIT(19)
#define PORTSC_OCC BIT(20)
#define PORTSC_PRC BIT(21)
#define PORTSC_PLC BIT(22)
#define PORTSC_CEC BIT(23)
#define PORTSC_CAS BIT(24)
#define PORTSC_WCE BIT(25)
#define PORTSC_WDE BIT(26)
#define PORTSC_WOE BIT(27)
#define PORTSC_DR BIT(30)
#define PORTSC_WPR BIT(31)

  uint32_t portpmsc;
  uint32_t portli;
  uint32_t reserved;
} __attribute__((packed));

/* 5.4 Host Controller Operational Registers
 * Table 26
 */
struct xhci_op_regs {
  uint32_t usbcmd;
#define XHCI_USBCMD_RS BIT(0)
#define XHCI_USBCMD_HCRST BIT(1)

  uint32_t usbsts;
#define XHCI_USBSTS_HCH BIT(0)
#define XHCI_USBSTS_CNR BIT(11)

  uint32_t pagesize;
  uint8_t reserved[8]; /* 0C - 13 */
  uint32_t dnctrl;     /* Device notification control */
  uint64_t crcr;       /* Command ring control */
#define XHCI_CRCR_CRP_MASK 0xFFFFFFFFFFFFFFC0
#define XHCI_CRCR_CRR BIT(3)
#define XHCI_CRCR_CRP_SIZE 4096

  uint8_t reserved1[16]; /* 20 - 2F */
  uint64_t dcbaap;       /* Device Context Base Address Array Pointer */
#define XHCI_DCBAAP_MAX_SIZE 2048

  uint32_t config; /* Configure */
#define XHCI_CONFIG_MAX_SLOT 44

  uint8_t reserved2[964]; /* 3C - 3FF */
                          /* USB Port register set */
#define XHCI_PORT_MAX 256
  struct xhci_port_regs prs[XHCI_PORT_MAX];
} __attribute__((packed));

/*
 * 5.5.2  Interrupter Register Set
 * Table 42: Interrupter Registers
 */
struct xhci_int_regs {
  uint32_t iman;
  uint32_t imod;
  uint32_t erstsz;
#define XHCI_ERST_SIZE_MASK 0xFFFF
  uint32_t reserved;
  uint64_t erstba;
#define XHCI_ERST_ADDR_MASK (~(0x3FUL))
  uint64_t erdp;
#define XHCI_ERDP_MASK (~(0xFUL))
} __attribute__((packed));

/* 5.5 Host Controller Runtime Registers */
struct xhci_run_regs {
  uint32_t mfindex; /* microframe index */
  uint8_t reserved[28];
#define XHCI_IRS_MAX 1024
  struct xhci_int_regs irs[XHCI_IRS_MAX];
} __attribute__((packed));

/* 5.6 Doorbell Registers*/
struct xhci_db_regs {
  uint8_t db[256];
} __attribute__((packed));